/*
  * Originates in the work for a seminar in the winter semester 2000/2001
  * at the University of Essen (http://www.uni-essen.de), department:
  * "MIS and software development" (http://nestroy.wi-inf.uni-essen.de).
  *
  * This DLL registers the function "BSF" with Rexx;
  * it allows Java to call into Rexx using the fine RexxSAA-interface
  * conventions and JNI ("Java Native Interface").
  *
  * version: 1.2.1 2003-01-06, ---rgf, finally found bug causing Object Rexx to bomb (but also Regina, it turned out!)
  *                               reason: freeing memory for Rexx-code with wrong method !
  *
  *                2003-01-05, ---rgf, adapted to Mark Hessling's great "Rexx/Trans"
  *                      (cf. <http://rexxtrans.sourceforge.net/>), which allows for
  *                      generic support for any supported Rexx interpreter, e.g.:
  *
  *                         Regina <http://regina-rexx.sourceforge.net/>,
  *                         IBM Object Rexx <http://www.software.ibm.com/ad/obj-rexx/>,
  *                         Quercus Systems Personal Rexx <http://www.quercus-sys.com/>,
  *                         Enterprise Rexx <http://www.WinREXX.com/>
  *
  *                 2001-06-11, ---rgf, runs with JNI 1 (does *not* need JRE 1.2 or higher anymore)
  *
  * author:  (c) 2001-2003 Rony G. Flatscher, WU (pronounced: "vey-u") Vienna University of Economics
  *                          and Business Administration
  *                       (preliminary work by Peter Kalender (Oct 2000 to Feb 2001), University of Essen)
  *
  *  LICENSE:
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
  *  License as published by the Free Software Foundation; either
  *  version 2.1 of the License, or (at your option) any later version.
  *
  *  This library is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  *  Lesser General Public License for more details.
  *
  *  You should have received a copy of the GNU Lesser General Public
  *  License along with this library; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
  *  The following URL may be available for further information:
  *
  *     <a href="http://www.gnu.org/copyleft/lesser.html">
  *              http://www.gnu.org/copyleft/lesser.html
  *     </a>

*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#define INCL_RXSHV
#define INCL_RXFUNC

// ---------------------- linking to which Rexx ----
#ifdef USE_REXXTRANS
# include <rexxtrans.h>

#elif defined USE_OREXX
# include <rexx.h>

#else
# include <rexxsaa.h>
#endif

#ifdef _MSC_VER
# pragma warning(disable:4100)
#endif


#ifdef OS2
   #define INCL_DOSMEMMGR
   #include <os2.h>
   #include <bsememf.h>
#endif

        // Java support
#include <jni.h>                        // "Java Native Interface" (JNI)
        // generated by Java to support the JNI
        // see comments at the end of "RexxAndJava.java"
#include "com_ibm_bsf_engines_rexx_RexxAndJava.h"



#ifdef __cplusplus
extern "C" {            // make sure no C++ name mangling comes into the way
#endif


#define RFH_BSF         "BSF"           // name of BSF function handler
#define DLLNAME         "BSF4Rexx"      // name of DLL and name for "PARSE SOURCE"

#define JAVA_CALL_BSF   "javaCallBSF"   // name of Java method to carry out the BSF-call



JNIEnv *globEnv;        // used by BSF() to call back to Java
jobject globObj;        // used by BSF() to call back to Java
jmethodID mid;          // used by BSF() to call back to Java



        // inline function for taking care of the different ways of freeing Rexx-allocated memory
inline void befreieSpeicher (void * memory)
{

#if defined (USE_REGINA) || defined (USE_REXXTRANS)     // 2003-01-04, ---rgf, added defined (USE_REXXTRANS)
//    printf(">>> befreieSpeicher(): 'USE_REGINA' or 'USE_REXXTRANS' defined, RexxFreeMemory()\n");
    RexxFreeMemory( memory );       // RexxAllocMemory( nrOfBytes );

#elif defined WIN32
//    printf(">>> befreieSpeicher(): 'WIN32' defined, GlobalFree()\n");
    GlobalFree( memory );          // GlobalAlloc( GMEM_FIXED, dwBytes nrOfBytes ); // only if GMEM_FIXED is a pointer returned
                                   //  HGLOBAL GlobalAlloc( UINT uFlags, DWORD dwBytes );
                                   // pv = GlobalAlloc(GMEM_FIXED, count);
                                   // WINBASE.H:#define GMEM_FIXED          0x0000


#elif defined OS2
//    printf(">>> befreieSpeicher(): 'OS2' defined, DosFreeMem()\n");
    DosFreeMem( memory );           // DosAllocMem(PVOID pointer, ULONG size, ULONG flags)
                                    // z.B. APIRET rc = DosAllocMem(&BaseAddress, ObjectSize, AllocationFlags);
                                    //    flags: PAG_READ (0x00000001), PAG_WRITE (0x00000002), PAG_EXECUTE (0x00000004)
                                    //     #define INCL_DOSMEMMGR
                                    //     #include <os2.h>
                                    //     #include <bsememf.h>

#else
//    printf(">>> befreieSpeicher():  using free()\n");
    free( memory );
#endif

}



/*
 * External Rexx function. This function allows Rexx programs to call into Java.
 */
APIRET APIENTRY
BSF  (  PSZ name,
        ULONG argc,
        PRXSTRING argv,
        PSZ queuename,
        PRXSTRING retstr)
     {

// printf(">>> BSF: arrived from Rexx, creating a Java array. \n");
        //convert arguments array to Java format
        jobjectArray arr = (*globEnv).NewObjectArray (
                                      argc,
                                    (*globEnv).FindClass("java/lang/String"),
                                    (*globEnv).NewStringUTF("")
                                                     );

// printf(">>> BSF: converting args to Java format. \n");
        // RXSTRINGs passed from Rexx have an additional \0 appended according Object Rexx 2.1 docs
        for(ULONG i=0; i<argc; i++)
        {
            if (argv[i].strptr==NULL)   // NULL (argument missing) ?
            {
               (*globEnv).SetObjectArrayElement(arr, i, NULL);
            }
            else                        // argument as a string in hand
            {
               (*globEnv).SetObjectArrayElement(arr, i,
                                               (*globEnv).NewStringUTF(argv[i].strptr));
            }
        }

// printf(">>> BSF: about to call Java method. \n");
        // call Java method and create Rexx return value
        jstring ReturnValue = (jstring)(*globEnv).CallObjectMethod(globObj, mid, arr);


// printf(">>> BSF: checking return value from Java method. \n");
        if (ReturnValue!=NULL)          // value returned from Java?
        {
                const char *str = globEnv->GetStringUTFChars(ReturnValue, 0);
                int         len = strlen(str);

                // o.k. not enough space available, free Rexx-allocated memory, re-allocate new chunk
                if (len>255)
                {
                   befreieSpeicher ( retstr->strptr );
                   retstr->strptr = new char[len+1];
                }

                strcpy(retstr->strptr, str);
                retstr->strlength = len;
                globEnv->ReleaseStringUTFChars(ReturnValue, str);
        }
        else    // o.k. Java returned null
        {
            // rgf, 2001-05
            if (retstr->strptr != NULL)
            {
                 befreieSpeicher ( retstr->strptr );
            }

            retstr->strptr    = NULL;   // set string pointer to NULL (as if no value was returned)
            retstr->strlength = 0;      // indicate no chars available
        }

                // pending exception ?
        if ( globEnv->ExceptionOccurred() != NULL ) // this works on JRE 1.1.8 ! ---rgf, 2001-05-25
        {
// printf(">>> BSF: ExceptionOccured in Java. \n");
             globEnv->ExceptionDescribe();      // print the stack
             globEnv->ExceptionClear();         // clear exception in the JVM

// printf("<<< BSF: returning to Rexx with error # 40. \n");
             return 40l;        // interrupt Rexx with "Incorrect call to routine!"
        }
// printf("<<< BSF: returning to Rexx, no error (everything o.k.). \n");
        return 0l;
     }



/*
 * This will be called from Java at initialization time and makes sure
 * that the external Rexx function named according to RFH_BSF will
 * registered. This way, Rexx programs being invoked from the Java side
 * have RFH_BSF available already (no need to register this function by
 * called Rexx program). Changed to "jint" on 2001-05-25.
 */
JNIEXPORT jint JNICALL
Java_com_ibm_bsf_engines_rexx_RexxAndJava_jniRegisterBSF (JNIEnv *env, jobject obj)
     {
         return (jint) RexxRegisterFunctionDll(RFH_BSF , DLLNAME, RFH_BSF );
     }


/*
 * This may be called from Java, pertains to Object Rexx only (as of 2003-01-04);
 * should wait until all cleanup within Object Rexx has been done. ---rgf, 2002-12-07, 2003-01-04
 */
JNIEXPORT jint JNICALL
Java_com_ibm_bsf_engines_rexx_RexxAndJava_jniRexxTerminate
  (JNIEnv *env, jobject obj)
     {
// printf(">>> JNI-jniRexxTerminate: just arrived from Java. \n");
#if defined (USE_OREXX) || defined( USE_REXXTRANS )
        RexxWaitForTermination();       // wait for Object Rexx to re-synch ...
// printf("\tBSF4Rexx.cc - Java_..._jniRexxTerminate(): just returned from that call.\n");
#endif
// printf("<<< JNI-jniRexxTerminate: returning to Java. \n");
        return (jint) 1;               // just return something

     }
  ;


/*
 * This is a pass-thru for RexxDidRexxTerminate(), returns 0 if Object Rexx still runs,
 * 1 if Object Rexx has terminated, ---rgf, 2002-12-13
 */
JNIEXPORT jint JNICALL
Java_com_ibm_bsf_engines_rexx_RexxAndJava_jniRexxDidRexxTerminate
  (JNIEnv *env, jobject obj)
     {
// printf(">>> JNI-jniRexxDidRexxTerminate: just arrived from Java. \n");
#if defined (USE_OREXX) || defined( USE_REXXTRANS )

// printf("<<< JNI-jniRexxTerminate: returning to Java by REALLY calling RexxDidRexxterminate(). \n");
        return (jint) RexxDidRexxTerminate();       // ask Object Rexx, if it is done

/*
int i= RexxDidRexxTerminate();
printf("\tBSF4Rexx.cc - Java_..._jniDidRexxTerminate(): i=%d.\n", i);
return (jint) i;
*/

#else
// printf("<<< JNI-jniRexxTerminate: returning to Java, fake call, always returns '1'. \n");
        return (jint) 1;        // default for non-Object Rexx is the integer value '1' for 'true'
#endif
     }
  ;



/*
 * This will be called from Java to invoke the Rexx interpreter. "source" contains
 * the Rexx program, "array" the arguments meant for the Rexx program.
 */
JNIEXPORT jstring JNICALL Java_com_ibm_bsf_engines_rexx_RexxAndJava_jniRexxStart
        ( JNIEnv *env,
          jobject         obj,
          jstring         source,
          jobjectArray    array)
     {
// printf("==> JNI-jniRexxStart: arrived from Java, will be calling Rexx. \n");
         // set global variables for callbacks to Java (via BSF())
         globEnv = env;
         globObj = obj;

                // get Java callback method via the Java object's class, looking up a method
                // named according to JAVA_CALL_BSF, signature: one argument (a String array) and
                // returning a String
         mid = (*globEnv).GetMethodID( (*globEnv).GetObjectClass(globObj),
                                       JAVA_CALL_BSF,
                                       "([Ljava/lang/String;)Ljava/lang/String;" );

         // make a copy of the Java string containing the Rexx script to be executed
         const char      *str = env->GetStringUTFChars(source, 0);
         char      *sourceStr = new char[strlen(str)+1];
         strcpy(sourceStr, str);
         env->ReleaseStringUTFChars(source, str);

         // create argument array for Rexx
         PRXSTRING argList = NULL;       // pointer to array of RXSTRINGS
         jsize   arrNum = 0;
         long    numArgs= 0l;

         if (array != NULL)              // an Object array available?
         {
             arrNum    =  env->GetArrayLength(array);
             numArgs   = (long) arrNum;
         }

         if (arrNum>0)
         {
            argList = new RXSTRING [ arrNum ];
            for (int i = 0; i < arrNum; ++i)
            {
               if ((env->GetObjectArrayElement(array, i))==NULL)/* no argument*/
               {
                  argList[i].strptr   =NULL;
                  argList[i].strlength=0;
               }
               else        /* a string argument in hand, give it to Rexx*/
               {
                  jstring        javaStr = (jstring) env->GetObjectArrayElement(array, i);
                  const char     *strArg = env->GetStringUTFChars(javaStr, 0);
                  int                len = strlen(strArg);
                  char        *tmpString = new char[len+1];
                  strcpy(tmpString, strArg);
                  env->ReleaseStringUTFChars(javaStr, strArg);

                  MAKERXSTRING( argList[i], tmpString, len);/* save arg for Rexx*/
               }
            }
         }

                // Rexx program to execute
         RXSTRING        RXsourceStr[2];
         MAKERXSTRING   (RXsourceStr[0], sourceStr, strlen(sourceStr));
         PRXSTRING       Instore = &RXsourceStr[0];
         MAKERXSTRING   (RXsourceStr[1], NULL, 0);

         SHORT           ReturnCode;
         LONG            ReturnValue;

                // let Rexx allocate buffer, if needed for a result
         RXSTRING        Result;
         MAKERXSTRING   (Result, NULL, 0);

         PSZ             ProgramName = DLLNAME ;

// printf("==> JNI-jniRexxStart: about to invoke Rexx via RexxStart(). \n");
         //start Rexx
         ReturnValue = RexxStart(
                         arrNum,         // LONG         ArgCount
                         argList,        // PRXSTRING    ArgList
                         ProgramName,    // PSZ          ProgramName
                         Instore,        // PRXSTRING    Instore
                         NULL,           // PSZ          EnvName
                         RXSUBROUTINE,   // LONG         CallType
                         NULL,           // PRXSYSEXIT   Exits
                         &ReturnCode,    // PUSHORT      ReturnCode
                         &Result         // PRXSTRING    Result
                         );

/* ---> ---rgf, 2003-01-05, seems that this is not needed by Object Rexx
#if defined (USE_OREXX) || defined( USE_REXXTRANS )     // 2003-01-04, ---rgf
        // as per 2001-06-11, hint from Mr. Micke about RexxWaitForTermination()

// printf("\tBSF4Rexx.cc - Java_..._jniRexxStart(): about to call RexxWatiForTermination()...\n");

// printf("==> JNI-jniRexxStart: RexxWaitForTermination(). \n");
        RexxWaitForTermination();       // wait for Object Rexx to re-synch ...

// printf("\tBSF4Rexx.cc - Java_..._jniRexxStart(): RexxWaitForTermination() just returned.\n");
#endif
<--- */

                // freeing memory for Rexx program and tokenized image
/*
printf("==> JNI-jniRexxStart: freeing Rexx program ... \n");
printf("                      strlength=%d ...", (int) RXsourceStr[0].strlength);
//        befreieSpeicher ( RXsourceStr[0].strptr );// free Rexx program ---> // 2003-01-05: this caused errors in Object Rexx (but also in Regina) !!!

        delete [] RXsourceStr[0].strptr;

printf(" o.k. \n");
*/

// printf("==> JNI-jniRexxStart: freeing tokenized image ... \n");
// printf("                      strlength=%d ...", (int) RXsourceStr[1].strlength);
        befreieSpeicher ( RXsourceStr[1].strptr );// free tokenized image of it
// printf(" o.k. \n");


        if (arrNum>0)   // free memory needed for Rexx arguments
        {
// printf("Java_..._jniRexxStart(): Freeing memory for Rexx arguments...\n");

// printf("==> JNI-jniRexxStart: freeing args ... \n");
           for (int i = 0; i < arrNum; ++i)
           {
// printf("                      #%d: strlength=%d ...", i, (int) argList[i].strlength);
              befreieSpeicher ( argList[i].strptr );
// printf(" o.k. \n");
           }
// printf("==> JNI-jniRexxStart: about to delete [] argList ...");
           delete [] argList;
// printf(" o.k. \n");
        }


        if (ReturnValue != 0l)     // error occurred, return null
        {
// printf("Java_..._jniRexxStart(): error occurred, freeing Rexx result-memory at #1 ...\n");

// printf("<== JNI-jniRexxStart: error occurred in Rexx, aborting & returning NULL. ReturnValue=%d ...", (int) ReturnValue);
           befreieSpeicher ( Result.strptr );     // free memory allocated from Rexx
// printf(" o.k. \n");
           return NULL;
        }

         // --- get result and return it to Java
        jstring jResult=NULL;           // define Java String to return

        if ( Result.strptr != NULL )    // a result was returned from Rexx
        {
             if (Result.strlength==0)   // empty string?
             {
                 jResult=env->NewStringUTF("");
             }
             else                       // valid non-empty string
             {
                 jResult=env->NewStringUTF( Result.strptr );
             }

// printf("Java_..._jniRexxStart(): freeing Rexx result-memory at #2 ...\n");

// printf("==> JNI-jniRexxStart: freeing 'Result.strptr' ... \n");
// printf("                      strlength=%d ...", (int) Result.strlength);
             befreieSpeicher ( Result.strptr ); // free memory where Rexx placed its result into
// printf(" o.k. \n");
        }

// printf("Java_..._jniRexxStart(): about returning jResult.\n");
// printf("<== JNI-jniRexxStart: ending returning jResult. \n");
        return jResult;         // return a Java String
     }



#ifdef __cplusplus
}               // closing bracket
#endif


