package com.ibm.bsf.engines.rexx;

/** Utility class to allow Rexx to address Java arrays, at() and put() implemented for up to
  * five dimensions for convenience. Generic implementation, that could be used by other
  * languages as well.
  *
  *<pre>
  *  AccessArrayUtil.java - convenience class for accessing up to 5-dimensional primitive Java arrays
  *  Copyright (C) 2001 Rony G. Flatscher
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
  *  License as published by the Free Software Foundation; either
  *  version 2.1 of the License, or (at your option) any later version.
  *
  *  This library is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  *  Lesser General Public License for more details.
  *
  *  You should have received a copy of the GNU Lesser General Public
  *  License along with this library; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
  *  The following URL may be available for further information:
  *
  *     <a href="http://www.gnu.org/copyleft/lesser.html">http://www.gnu.org/copyleft/lesser.html</a>
  *</pre>
  *
  * @version 1.0.0, 2001-06-02
  * @author Rony G. Flatscher (WU Wien University)
  */
public abstract class AccessArrayUtil
{
   /** Get an element of a one dimensional array. */
   static public Object at(Object [] array, int i1)
   {
      return array [i1];
   }

   /** Get an element of a two dimensional array. */
   static public Object at(Object [][] array, int i1, int i2)
   {
      return array [i1] [i2];
   }

   /** Get an element of a three dimensional array. */
   static public Object at(Object [] [] [] array, int i1, int i2, int i3)
   {
      return array [i1] [i2] [i3];
   }

   /** Get an element of a four dimensional array. */
   static public Object at(Object [] [] [] [] array, int i1, int i2, int i3, int i4)
   {
      return array [i1] [i2] [i3] [i4];
   }

   /** Get an element of a five dimensional array. */
   static public Object at(Object [] [] [] [] [] array, int i1, int i2, int i3, int i4, int i5)
   {
      return array [i1] [i2] [i3] [i4] [i5];
   }


   /** Put an element into a one dimensional array. */
   static public void put(Object [] array, Object newValue, int i1)
   {
       array [i1] = newValue;
   }

   /** Put an element into a two dimensional array. */
   static public void put(Object [][] array, Object newValue, int i1, int i2)
   {
       array [i1] [i2] = newValue;
   }

   /** Put an element into a three dimensional array. */
   static public void put(Object [] [] [] array, Object newValue, int i1, int i2, int i3)
   {
       array [i1] [i2] [i3] = newValue;
   }

   /** Put an element into a four dimensional array. */
   static public void put(Object [] [] [] [] array, Object newValue, int i1, int i2, int i3, int i4)
   {
       array [i1] [i2] [i3] [i4] = newValue;
   }

   /** Put an element into a five dimensional array. */
   static public void put(Object [] [] [] [] [] array, Object newValue, int i1, int i2, int i3, int i4, int i5)
   {
       array [i1] [i2] [i3] [i4] [i5] = newValue;
   }
}

